from ladim_aggregate import geotag
import json
import xarray as xr
import pkgutil


class Test_create_geotagger:
    def test_can_return_correct_polygon_attribute_of_particle(self):
        chunk = xr.Dataset(
            data_vars=dict(
                lon=xr.Variable('pid', [.5, .5, 10.5]),
                lat=xr.Variable('pid', [60.5, 70.5, 70.5]),
            )
        )

        pkg = 'ladim_aggregate.examples.connect'
        geojson = json.loads(pkgutil.get_data(pkg, 'regions.geojson').decode('utf-8'))

        geotagger = geotag.create_geotagger(
            attribute="region",
            x_var='lon',
            y_var='lat',
            geojson=geojson,
            missing=-1,
        )

        region = geotagger(chunk)
        assert region.dims == ('pid', )
        assert region.values.tolist() == [101, -1, 102]


class Test_lookup:
    def test_correct_when_1d_indicator_array(self):
        indicator = [False, True, False]
        values = ['A', 'B', 'C']
        result = geotag.lookup(indicator, values, None)
        assert result == 'B'

    def test_returns_missing_value_when_no_indicator(self):
        indicator = [False, False, False]
        values = ['A', 'B', 'C']
        result = geotag.lookup(indicator, values, 'D')
        assert result == 'D'

    def test_correct_when_2d_indicator_array(self):
        indicator = [
            [False, True],
            [False, False],
            [True, False],
        ]
        values = ['A', 'B', 'C']
        result = geotag.lookup(indicator, values, None)
        assert result.tolist() == ['C', 'A']


class Test_get_tag:
    def test_correct_when_standard_input(self):
        result = geotag.get_tag(
            xy=[(1, 1), (3, 3), (5, 5)],
            polygons=[
                geotag.Polygon(shell=[(0, 0), (2, 0), (2, 2), (0, 2)], holes=[]),
                geotag.Polygon(shell=[(2, 2), (4, 2), (4, 4), (2, 4)], holes=[]),
            ],
            attributes=[101, 202],
            missing=404,
        )
        assert list(result) == [101, 202, 404]


class Test_polygons_contains_points:
    def test_correct_when_simple_polygon(self):
        poly = geotag.Polygon(shell=[(10, 0), (12, 0), (12, 2), (10, 2)], holes=[])
        xy = [(11, 1), (13, 3)]
        result = poly.contains_points(xy)
        assert list(result) == [True, False]

    def test_returns_boolean_value_when_boundary(self):
        poly = geotag.Polygon(shell=[(0, 0), (2, 0), (2, 2), (0, 2)], holes=[])
        xy = [(0, 0), (2, 2), (1, 2), (0, 1)]
        result = poly.contains_points(xy)
        assert set(result) - {True, False} == set()

    def test_correct_when_holes(self):
        poly = geotag.Polygon(
            shell=[(10, 0), (15, 0), (15, 5), (10, 5)],
            holes=[
                [(11, 1), (13, 1), (13, 3), (11, 3)]
            ],
        )
        xy = [(12, 2), (14, 4), (16, 6)]
        result = poly.contains_points(xy)
        assert list(result) == [False, True, False]


class Test_Polygon_from_geojson_feature:
    def test_correct_when_no_holes(self):
        coords = [[[0, 0], [2, 0], [2, 2], [0, 2]]]
        f = {"geometry": {"coordinates": coords}}
        poly = geotag.Polygon.from_geojson_feature(f)
        assert poly.shell.vertices.tolist() == coords[0]
        assert len(poly.holes) == 0

    def test_correct_when_holes(self):
        shell = [[0, 0], [3, 0], [3, 3], [0, 3]]
        hole = [[1, 1], [2, 1], [2, 2]]
        coords = [shell, hole]
        f = {"geometry": {"coordinates": coords}}
        poly = geotag.Polygon.from_geojson_feature(f)
        assert poly.shell.vertices.tolist() == shell
        assert len(poly.holes) == 1
        assert poly.holes[0].vertices.tolist() == hole



import xarray as xr
from matplotlib.path import Path
import numpy as np


def create_geotagger(attribute, x_var, y_var, geojson, missing=np.nan):
    """
    Create geotagger function.

    :param attribute: Polygon attribute to return
    :param x_var: Name of x coordinate variable
    :param y_var: Name of y coordinate variable
    :param geojson: A dict representation of a geojson file
    :param missing: Value to return if coordinates does not match a polygon
    :return: A geotagger function
    """
    props = attributes_from_geojson(geojson, attribute)
    paths = paths_from_geojson(geojson)

    def geotagger(chunk):
        """
        Returns attributes based on geographical position

        :param chunk: An xarray dataset containing coordinates
        :return: An xarray variable with attributes from enclosing polygons
        """
        x = chunk[x_var].values
        y = chunk[y_var].values
        xy = np.stack([x, y]).T
        point_props = get_tag(xy, paths, props, missing)
        return xr.Variable(dims='pid', data=point_props)

    return geotagger


def attributes_from_geojson(geojson: dict, attribute) -> list:
    """
    Extract attributes from geojson dict

    :param geojson: A geojson dict
    :param attribute: Name of attribute to return
    :return: A list of attribute values
    """
    return [f['properties'][attribute] for f in geojson['features']]


class Polygon:
    def __init__(self, shell, holes):
        holes = holes or []

        self.shell = Path(vertices=shell)
        self.holes = [Path(vertices=h) for h in holes]

    @staticmethod
    def from_geojson_feature(feature):
        coords = feature['geometry']['coordinates']
        return Polygon(shell=coords[0], holes=coords[1:])

    def contains_points(self, xy) -> np.ndarray:
        """
        For each input point, return True (if inside) or False (if not inside)

        :param xy: Input X/Y coordinates
        :return: An array of true/false values
        """
        result = self.shell.contains_points(xy)
        for hole_path in self.holes:
            result = result & (~hole_path.contains_points(xy))
        return result


def paths_from_geojson(geojson: dict) -> list[Polygon]:
    """
    Convert a geojson dict to matplotlib Path objects

    :param geojson: A geojson dict
    :return: A list of matplotlib paths
    """
    features = geojson['features']
    polys = []
    for feature in features:
        polys.append(Polygon.from_geojson_feature(feature))

    return polys


def lookup(indicator, values, missing):
    """
    Use an indicator array to return values from a value array.

    :param indicator: Indicator array (true/false), shape N x M
    :param values: Value array (return first value that corresponds to an
        indicator value of True), shape N
    :param missing: Value to return if entire indicator array is False
    :return: The first corresponding value from the values array, or `missing`
        if the entire indicator value is False, shape M
    """
    np_values = np.asarray(list(values) + [missing])
    first_nonzero = np.sum(np.cumsum(indicator, axis=0) == 0, axis=0)
    return np_values[first_nonzero]


def get_tag(xy, polygons: list[Polygon], attributes: list, missing):
    """
    Return a polygon attribute for each input point

    For each point, find the first polygon that contains the point and
    return the corresponding attribute. If no polygon contains the point,
    return the `missing` value.

    :param xy: Coordinates of points (shape N x 2)
    :param polygons: List of polygons
    :param attributes: One attribute for each polygon
    :param missing: The value to return if some points are not in the polygons
    :return: A list of polygon attributes, one for each point
    """
    inside = np.asarray([p.contains_points(xy) for p in polygons])
    return lookup(inside, attributes, missing)
